/*	System I/O routines for scan300c

Copyright (c) Terence P. Gliedt, 1992.
	This program may be freely redistributable in its original, unmodified form.
	The author grants you a nonexclusive, royalty-free license to use this software,
	and modify it as needed provided that this copyright notice remains, and any
	changes are clearly indicated as such.  No part may be incorporated in products or
	packages sold for profit without express permission of the copyright holder.
	The author assumes no responsibility for the consequences of using this program
	and no warranty of any sort is expressed or implied.
*/
#include <stdio.h>
#include <sys/file.h>
#include <signal.h>
#include "es300c.h"
#include <nextdev/scsireg.h>
#include <sgtty.h>

#define TIMEOUT 10		/* Number seconds to wait for scanner to respond */

/*
DoRead (Here,  Length)

Function:		Read 'Length' bytes into Here

Returns:		Length of data read or 0 if read failed.
*/
int DoRead (char *Here, int Length)
{
	struct scsi_req sr;				/* Used with ioctl SGIOCREQ */
	struct cdb_6s* cdbp = &sr.sr_cdb.cdb_c6s;
	int Total;
	int Residual;	
	int Bytes;

	/************************************************************
		Read	Serial/Parallel	Device	here
	************************************************************/
	if (ScannerType != SCSI) {
		for(Total=0, Residual=Length; Residual; Total=Total+Bytes) {
			Bytes = read(Scanner, Here, Residual);
			if (Bytes <= 0) {
				fprintf(stderr, "%s: Error reading %d bytes\n", Length);
				perror("Read error");
				return(Total);
			}
			Residual = Residual - Bytes;
			Here = Here + Bytes;
		}
		XDEBUG2("\tRead issued for %d bytes. Read %d bytes\n", Length, Total);
		if (Debug>1) {
			if (*Here == ACK) fprintf(stderr, "\tReceived ACK\n");
			else if (*Here == STX) fprintf(stderr, "\tReceived STX\n");
			else if (*Here == NAK) fprintf(stderr, "\tReceived NAK\n");
			else fprintf(stderr,"\tReceived data '%2.2xX' from scanner\n", (int) *Here);
		}
		return(Total);
	}

	/************************************************************
		Read	SCSI	Device	here
	************************************************************/
	bzero(&sr, sizeof(struct scsi_req));
	cdbp->c6s_opcode = C6OP_READ;
	cdbp->c6s_lun = 0;
	cdbp->c6s_len = Length;			/* Number of bytes to device */
	sr.sr_dma_dir = SR_DMA_RD;
	sr.sr_addr = Here;				/* Read into here */
	sr.sr_dma_max = Length;		/* Number of bytes to device */
	sr.sr_ioto	= TIMEOUT;			/* I/O timeout in seconds */
	if (DoIoctl(Scanner, ScannerDevice, &sr)) return(0);
	XDEBUG3("\tRead issued. Status=%d, SCSI status byte=%2.2x Number bytes=%d\n",
		sr.sr_io_status, sr.sr_scsi_status, sr.sr_dma_xfr);
	if (Debug>1) {
		if (*Here == ACK) fprintf(stderr, "\tReceived ACK\n");
		else if (*Here == STX) fprintf(stderr, "\tReceived STX\n");
		else if (*Here == NAK) fprintf(stderr, "\tReceived NAK\n");
		else fprintf(stderr,"\tReceived data '%2.2xX' from scanner\n", (int) *Here);
	}
	return(Length);
}

/*
DoWrite (Here,  Length)

Function:		Write 'Length' bytes from Here

Returns:		Length of data written or 0 if it failed.
*/
int DoWrite (char *Here, int Length)
{
	struct scsi_req sr;				/* Used with ioctl SGIOCREQ */
	struct cdb_6s* cdbp = &sr.sr_cdb.cdb_c6s;
	int Bytes;

	/************************************************************
		Write	Serial/Parallel	Device	here
	************************************************************/
	if (ScannerType != SCSI) {
		Bytes = write(Scanner, Here, Length);
		if (Bytes <= 0) {
			fprintf(stderr, "%s: Error writing %d bytes\n", Length);
			perror("Write error");
		}
		XDEBUG2("\tWrite issued for %d bytes. Wrote %d bytes\n", Length, Bytes);
		return(Bytes);
	}

	/************************************************************
		Write	SCSI	Device	here
	************************************************************/
	/* Fill out block to talk directly to device */
	bzero(&sr, sizeof(struct scsi_req));
	cdbp->c6s_opcode = C6OP_WRITE;
	cdbp->c6s_lun = 0;
	cdbp->c6s_len = Length;			/* Number of bytes for device */
	sr.sr_dma_dir = SR_DMA_WR;
	sr.sr_addr = Here;				/* Here is data to write */
	sr.sr_dma_max = Length;		/* Number of bytes for device */
	sr.sr_ioto	= TIMEOUT;			/* I/O timeout in seconds */
	if (DoIoctl(Scanner, ScannerDevice, &sr)) return(0);
	XDEBUG3("\tWrite issued. Status=%d SCSI status byte=%2.2x Number bytes=%d\n",
		sr.sr_io_status, sr.sr_scsi_status, sr.sr_dma_xfr);
	return(Length);
}

/*
FindScanner(ScannerDev, Protocol, Target, Speed)

Function:		Open the device ScannerDev, setting Scanner.  Protocol may be used
			to force how the device should be driven (serial, scsi, parallel).
			This may also be NULL meaning we should figure it out ourselves.

			If this is a SCSI device, and Target is zero, inquire on the SCSI bus
			for the scanner. Otherwise check that Target is a scanner.
			If this is a serial device, open the device for a speed of Speed (parity=None).

Returns:		-1 if no scanner found, >= 0 if scanner was found
*/
int FindScanner(char *ScannerDev, char *Protocol, int T, char *Speed)
{
	struct scsi_adr sa;
	struct scsi_req sr;
	struct cdb_6 *cdbp = &sr.sr_cdb.cdb_c6;
	char InquiryData[64];		/* Read Inquiry data into this */
	int Tstart =0;				/* Try targets in this range */
	int Tend = 7;

	struct sgttyb tty;			/* Used to set characteristics of serial line. */
	int LineDiscipline;
	
	/* Figure out the type of device. SCSI is easy. Anything else is serial on NeXT. */
	if (!Protocol) {
		if (!strcmp(ScannerDev, "/dev/sg0")) ScannerType = SCSI;
		else ScannerType = SERIAL;
	}
	else {
		ScannerType = 0;
		/* User specified a protocol. See if he can spell */
		if (!strcmp(Protocol, "serial")) ScannerType = SERIAL;
		if (!strcmp(Protocol, "scsi")) ScannerType = SCSI;
		if (!strcmp(Protocol, "parallel")) ScannerType = PARALLEL;
		if (!ScannerType) {
			fprintf(stderr, "%s: Invalid scanner protocol '%s' for device '%s'\n",
				MeName, Protocol, ScannerDev);
			return(-1);
		}
	}

	/* Attempt to open the scanner. Scanner is global. If this doesn't work, all is for naught. */
	if ((Scanner = open(ScannerDev, O_RDWR, 0666)) <= 0) {
		fprintf(stderr, "%s: Unable to open scanner device '%s'\n", MeName, ScannerDev);
		return(-1);
	}

	/************************************************************
		Serial	Device	here
	************************************************************/
	if (ScannerType == SERIAL) {	
		tty.sg_ispeed = EXTA;
		if (!strcmp(Speed, "9600")) tty.sg_ispeed = B9600;
		tty.sg_ospeed = tty.sg_ispeed;
		tty.sg_flags = RAW;
		/* Set system treatment of device */
		if (ioctl(Scanner, TIOCSETP, &tty)>0) {
			fprintf(stderr,"%s: Unable to set serial line %s to speed=%02x  flags=%02x\n",
				MeName, ScannerDev, tty.sg_ispeed, tty.sg_flags);
			return(-1);					/* No device found */
		}
		XDEBUG3("\tSerial device %s openned. Speed=%02x, flags=%02x\n",
			ScannerDev, tty.sg_ispeed, tty.sg_flags);
		return(0);						/* Device openned */
	}

	/************************************************************
		Parallel	Device	here
	************************************************************/
	if (ScannerType == PARALLEL) {	
		XDEBUG("\tSerial device %s openned. \n", ScannerDev);
		return(0);						/* Device openned */
	}

	/************************************************************
		SCSI	Device	here
	************************************************************/
	/* If Target is already selected, just verify its there etc. */
	if (T) Tstart = Tend = T;

	/* Search all targets on unit 0. If the SCSI inquiry returns a string that identifies an
		EPSON device,  Thats the one we use. */
	for(T=Tstart; T<= Tend; T++) {
		sa.sa_target = T;
		sa.sa_lun = 0;
		if (ioctl(Scanner, SGIOCSTL, &sa) >= 0) {
			bzero(&sr, sizeof(struct scsi_req));
			cdbp->c6_opcode = C6OP_INQUIRY;
			cdbp->c6_lun = 0;
			cdbp->c6_lba = 0;				/* Start block for read/write */
			cdbp->c6_len = sizeof(InquiryData) ;  /* Number of bytes from device */
			sr.sr_dma_dir = SR_DMA_RD;
			sr.sr_addr = (char *) &InquiryData;	/* Read into here */
			sr.sr_dma_max = cdbp->c6_len;	/* Number of bytes from device */
			sr.sr_ioto	=2;					/* I/O timeout in seconds */
			if (ioctl(Scanner, SGIOCREQ, &sr) >= 0)  {
				if(!sr.sr_io_status) {
					XDEBUG2("\tTarget=%d Data='%20.20s'\n", T, &InquiryData[8]);
					if (!strncmp("EPSON", &InquiryData[8]) ) {
						Target = T;		/* Save global value */
						XDEBUG("\tAssigned target=%d\n", Target);
						return(T);
					}
				}
			}
		}
	}
	fprintf(stderr,"%s: Unable to find Epson SCSI device\n", MeName);
	return(-1);					/* No device found */
}

/*
DumpScsi_Req (Request, String)

Function:		Dump a SCSI_REQ block. String is put in the header.

Returns:		none
*/
void DumpScsi_Req(struct scsi_req *Request, char * String)
{
	fprintf(stderr, "%s - Dump of scsi_req:\n", String);
	fprintf(stderr, "\tCDB=%s\n", Hex(Request, sizeof(struct cdb_6)));
	fprintf(stderr, "\tsr_dma_dir=%d\n", Request->sr_dma_dir);
	fprintf(stderr, "\tsr_addr=%8.8x\n", Request->sr_addr);
	fprintf(stderr, "\tsr_dma_max=%d\n", Request->sr_dma_max);
	fprintf(stderr, "\tsr_ioto=%d\n", Request->sr_ioto);
	fprintf(stderr, "\tsr_io_status=%d\n", Request->sr_io_status);
	fprintf(stderr, "\tsr_scsi_status=%02x\n", Request->sr_scsi_status);
	fprintf(stderr, "\tsr_esense=%s\n", Hex(&(Request->sr_esense), sizeof (struct esense_reply)));
	fprintf(stderr, "\tsr_dma_xfr=%d\n", Request->sr_dma_xfr);
	fprintf(stderr, "\tsr_exec_time=%s\n",  Hex(&(Request->sr_exec_time), sizeof(struct timeval)));
	return;
}
	
/*
DoSense ()

Function:		Issue sense to device for logical unit LogicalUnit

Returns:		0 if successful, 1 if not
*/
int DoSense ()
{
	struct scsi_req sr;				/* Used with ioctl SGIOCREQ */
	struct cdb_6* cdbp = &sr.sr_cdb.cdb_c6;
	char SenseInfo[24];			/* Read sense information here */

	/************************************************************
		Serial/Parallel	Device	here
	************************************************************/
	if (ScannerType != SCSI) {	
		XDEBUG("\tNo sense info for non-SCSI device '%s'\n", ScannerDevice);
		return(0);
	}
	
	/************************************************************
		SCSI	Device	here
	************************************************************/
	/* Issue sense to clear unit attention */
	bzero(&sr, sizeof(struct scsi_req));
	cdbp->c6_opcode = C6OP_REQSENSE;
	cdbp->c6_lun = LogicalUnit;
	cdbp->c6_len = sizeof(SenseInfo) ;
	sr.sr_dma_dir = SR_DMA_RD;
	sr.sr_addr = SenseInfo;			/* Read information here */
	sr.sr_dma_max = sizeof(SenseInfo) ;	/* Max # bytes to transfer */
	sr.sr_ioto	= TIMEOUT;			/* I/O timeout in seconds */
	if (!DoIoctl(Scanner, ScannerDevice, &sr)) {
		XDEBUG3("\tSense issued. Status=%d SCSI status byte=%02x Number bytes=%d\n",
			sr.sr_io_status, sr.sr_scsi_status, sr.sr_dma_xfr);
		XDEBUG("\tSense info=%s\n", Hex((char *) SenseInfo, 8));
	}
	else return(1);				/* Eeek - device failed even sense ! */
	return(0);
}

/*
DoIoctl (Device, DeviceName, Request)

Function:		Issue ioctl to Device (whose name is DeviceName). The request
			block is Request and has been filled out by the caller.

Returns:		0 or 1 depending whether the request failed or succeeded
*/
int DoIoctl(int Device, char *DeviceName, struct scsi_req *Request)
{
	int I;
	const char *IOStatusMsgs[18] = {	/* Values from /usr/include/nextdev/scsireg.h */
		"Successful",				/* SR_IOST_GOOD */
		"Selection timeout",			/* SR_IOST_SELTO */
		"Check status, sr_esense valid",	/* SR_IOST_CHKSV */
		"Tried to move > sr_dma_max",	/* SR_IOST_DMAOR */
		"Sr_ioto exceeded",			/* SR_IOST_IOTO */
		"Bus violation",				/* SR_IOST_BV */
		"Command reject",			/* SR_IOST_CMDREJ */
		"Memory allocation failure",	/* SR_IOST_MEMALL */
		"Memory fault",			/* SR_IOST_MEMF */
		"Not super user",			/* SR_IOST_PERM */
		"Device not open",			/* SR_IOST_NOPEN */
		"Target aborted command",	/* SR_IOST_TABT */
		"Bad SCSI status byte",		/* ST_IOST_BADST */
		"Internal driver error",		/* ST_IOST_INT */
		"Unexpected byte count",		/* SR_IOST_BCOUNT */
		"Desired volume not available",	/* SR_IOST_VOLNA */
		"Media Write Protected",	/* SR_IOST_WP */
		"?"
	};
	const char *SenseMsgs[9] = {		/* Values from /usr/include/nextdev/scsireg.h */
		"No sense to report",		/* SENSE_NOSENSE */
		"Recovered error",			/* SENSE_RECOVERED */
		"Not ready",				/* SENSE_NOTREADY */
		"Media flaw",				/* SENSE_MEDIA */
		"H/W failure",				/* SENSE_HARDWARE */
		"Illegal request",			/*  */
		"Unit attention",			/* SENSE_UNITATTENTION */
		"Data access protected",		/* SENSE_DATAPROTECT */
		"?"
	};
	
	if (ioctl(Device,SGIOCREQ, Request) < 0) {
		fprintf(stderr,"%s: Error issuing ioctl to device %s. Error number=%d\n",
			MeName, DeviceName, errno);
		perror("ioctl(SGIOCREQ)");
		return(1);
	}
	if (Request->sr_io_status == SR_IOST_GOOD) return(0);   /* No problemo */
				
	/* Ooops - ioctl failed. Report it */
	I = Request->sr_io_status;
	if (I >18-1) I = 18-1;
	fprintf(stderr, "%s: sr_io_status=%02XX (%s)\n",
		MeName, Request->sr_io_status, IOStatusMsgs[I]);
	if(Request->sr_io_status == SR_IOST_CHKSV) {
		I = Request->sr_esense.er_sensekey;
		if (I >9-1) I = 9-1;
		fprintf(stderr, " \tSense key=%02XX  (%s) Sense code = %02XX\n",
			Request->sr_esense.er_sensekey, SenseMsgs[I], 			Request->sr_esense.er_addsensecode);
	}
	if (Debug) DumpScsi_Req(Request, "Block in error follows");
	return(1);
}
