#include <stdio.h>
#define MAIN			/* Be sure global data is defined here */
#include "es300c.h"

int DoScan();			/* Scan data from scanner */
int CmdNoParms();		/* Execute scanner command with no parms */
int CmdWithParms();	/* Execute scanner command with parm */
int CmdReqData();		/* Execute scanner command, read in data*/
int AckNak();			/* Read an ACK or NAK from scanner */
int ByteCounter ();		/* Convert 2-byte byte counter to host format */
void MakeByteCounter();		/* Create 2-byte byte counter for scanner */
void Usage();			/* Short help for user */

#define VERSION "0.7"
/*
scan300c 
 
Function:    Drive a color scanner. This program is not intended to have
	a good user interface nor to guide a person in knowing just what parameters to use
	to use the scanner. This is rather a low-level pgm to be driven by another program
	which really does understand the nuances of how the particular scanner works.
	
	At the moment only the Epson 300C scanner is supported.

Syntax:
	{scanner-name}  [-options]   outfile
 
	outfile	is the file to be created. If not specified, data is written to stdout
 
Options:   (controlled by [-options] in the command)

	a MO SO ML SL	Set reading area.  offset=MO,SO area=ML, SL
	b N		Serial line speed. N may be 9600 or 19200
	c		Set to color scan
	d dev	Device to open. Default is DEVICE
	f target	NeXT SCSI device address. Default is to search for device.
	g		Set to gray-scale scan.  8 bits/pixel
	g4		Set to gray-scale scan.  4 bits/pixel
	g2		Set to gray-scale scan.  2 bits/pixel
	i		Initialize
	m		Set to monochrome scan
	n name	Name of document
	ppm		Generate the data in PPM format (unimplemented)
	q		Query status of device
	r N		Set resolution to N dpi (should be done before -A)
	s File	Scan at current settings to File. PPM files may be '-'
	tiff		Generate the data in TIFF format (unimplemented)
 	v		Verbose level
	x		Quiet mode, no warm fuzzy msgs
	y serial/scsi/parallel		Force device protocol 
 	z N		Set zoom level to N (should be done before -A)
 
 	Options corresponding exactly to 300C commands.  Little error checking is
 	done here on any parameters. You can mess up the scanner using these.
	
	I		Identify cmd
	F		Status cmd
	S		Condition cmd
	G		Start scanning
	C xx		Set  color mode to xx (0, 10, 20, 30, 1, 2)
	D n		Set data format  (1-8) 
	R M,S	Set main scan resolution to M, Subscan to S
	H M,S	Set main scan zoom to M, Subscan to S (50-200)
	A MO SO ML SL	Set reading area.  offset=MO,SO area=ML, SL
	L xx		Set brightness to xx (3, 2, 1, 0, ff, fe, fd)
	Z xx		Set gamma to xx (1, 2, 0, 10, 20)
	B xx		Set halt tone to xx (0, 10, 20, 01)
	M xx		Set color correction to xx (0, 20, 40, 80)
	@		Initialize

Performance:
	The Epson scanner does not deliver data as fast as one might want. This
	seems to be due, at least in part, to a reportedly slow SCSI bus in the NeXT.
	Stay tuned to see how this performs on the RISC/6000.
	
Bugs:
	Epson scanner h/w provides no means to reset when scanner is in E-1 state.
	Add support for non-NeXT platforms

Note:
	The device used by this program must have its access bits set properly 
	to allow the general user to read & write to it. This is normally not a problem.
	
	All rights to this code were relinquished by IBM Rochester IPL in Aug 1992.

References:
	"Epson Color-Monochrome Image Scanners ES-300C Programmers Guide"
		(X-ES300C-P) available from Epson at 1-800-922-8911 for $7.50.
	Chapter 4 "Unix Programmer's manual Reference Guide" by Computer Systems
		 Research Group, University of California, Berkely, Ca.

Thanks:
	To NeXT for its Generic SCSI driver example in /NextDeveloper/Examples/SCSI.
	Thanks to Sam Leffler for his TIFF Software "Release" 3.00.
	Thanks to Jef Poskanzer jef@well.sf.ca.us for his PBMplus software.

Copyright (c) Terence P. Gliedt, 1992.
	This program may be freely redistributable in its original, unmodified form.
	The author grants you a nonexclusive, royalty-free license to use this software,
	and modify it as needed provided that this copyright notice remains, and any
	changes are clearly indicated as such.  No part may be incorporated in products or
	packages sold for profit without express permission of the copyright holder.
	The author assumes no responsibility for the consequences of using this program
	and no warranty of any sort is expressed or implied.
*/
int  main ( int argc, char *argv[] )
{
	char  CmdList[100];		/* Sequence of unique commands to execute */
	int     CmdIndex = 0;	/* Which command we are going to execute */
	char  Cmd  = C_None;	/* This is the command to do */
	char *LineSpeed = "19200";		/* Speed of line: 9600, 19200 */
 
	int	   I,J,K;			/* Bean counters */
	short   shortI;			/* Used to make machine normal 2-byte integers */
	char    *Res;			/* Bean pointer */
	int       HexInt;			/* Integer from hexadecimal input */
	u_char HexByte;		/* One byte of data */
	u_char HexBytes[2];	/* Two one-byte chunks of data */
	struct {				/* Keep these all together */
		u_char HexShort[2];		/* Two bytes of data */
		u_char HexShort1[2];
		u_char HexShort2[2];
		u_char HexShort3[2];
	} HexParm;

	/* Read data into these areas */
	struct 	Identify_Data_Block 	ScannerIdentifyDB;
	struct 	Status_Data_Block 		ScannerStatusDB;
	struct 	Condition_Data_Block 	ScannerConditionDB;
	
	/* Data from saved from various options */
	char *Protocol = NULL;		/* How to drive device, serial, scsi or parallel */
	char *DocName = NULL;		/* Name of scanned in document */
	char *FileName = "stdout";	/* Name of file we are writing to */
	char 	*ColorMode = "0";		/* From -C */
	char 	*DataFormat = "1";		/* From -D */
	char 	*Resolution1 = "100";	/* From -R */
	char *Resolution2 = "100";
	char *Zoom1= "100";		/* From -H */
	char *Zoom2 = "100";
	char *ReadingArea1 = "0";	/* From -A */
	char *ReadingArea2 = "0";
	char *ReadingArea3 = "0";
	char *ReadingArea4 = "0";
	char *Brightness = "0";		/* From -L */
	char *GammaCorrection = "1";	/* From -Z */
	char *HalfToningMode = "0";	/* From -B */
	char *ColorCorrection = "80";	/* From -M */

	/* Right off, check if NO parms were specified. If so, show help and exit */
	if (argc < 2) Usage();
	
	/*	Parse options. Note this technique requires parameters for options to be SEPARATE
		words. I.e. "-d /dev/sg0" is not the same as "-d/dev/sg0".  */
	ScannerIdentifyDB.identification[1] = ' ';	/* So we can tell Identify was done or not */
	while((argc>1) && (argv[1][0] == '-')) {
		for (I=1; I < strlen(argv[1]); I++) {
			switch (argv[1][I]) {
			/* These are options to the pgm */
			case 'b' :
				if (argc > 2) {
					LineSpeed = argv[2];
					I = 255; argc--;  argv++;
				}
				break;
			case 'c' :
				CmdList[CmdIndex] =C_Set_Color_Mode;
				CmdIndex++;
				ColorMode = NULL;		/* Means you hafta figure it out */
				CmdList[CmdIndex] =C_Set_Data_Format;
				CmdIndex++;
				DataFormat = "8";		/* Bits per pixel */
				break;
			case 'd' :
				if (argc > 2) {
					ScannerDevice = argv[2];
					I = 255; argc--;  argv++;
				}
				break;
			case 'f' :
				if (argc > 2) {
					sscanf(argv[2], "%d", &Target);
					I = 255; argc--;  argv++;
					DEBUG("Using device at SCSI address %d\n", Target);
				}
				break;
			case 'g' :
				CmdList[CmdIndex] =C_Set_Color_Mode;
				CmdIndex++;
				ColorMode = "0";
				CmdList[CmdIndex] =C_Set_Data_Format;
				CmdIndex++;
				DataFormat = "8";		/* Set default value */
				if (argv[1][I+1] == '4') { I++; DataFormat = "4"; }
				if (argv[1][I+1] == '2') { I++; DataFormat = "2"; }
				break;
			case 'i' :
				CmdList[CmdIndex] =C_Initialize;
				CmdIndex++;
				break;
			case 'm' :
				CmdList[CmdIndex] =C_Set_Color_Mode;
				CmdIndex++;
				ColorMode = "0";
				CmdList[CmdIndex] =C_Set_Data_Format;
				CmdIndex++;
				DataFormat = "1";		/* Bits per pixel */
				break;
			case 'n' :
				if (argc > 2) {
					DocName = argv[2];
					I = 255; argc--;  argv++;
				}
				break;
			case 'r' :
				CmdList[CmdIndex] =C_Set_Resolution;
				CmdIndex++;
				if (argc > 2) {
					Resolution1 = argv[2];
					Resolution2 = argv[2];
					I = 255; argc--;  argv++;
				}
				/* Check all parms before this. If any -A command is found,
				 	force the -A to come AFTER this command, as this command
					causes a reset of the area to its maximum value. */
				K = CmdIndex;			/* Only look so far */
				for(J=0; J<K; J++) {
					if (CmdList[J]==C_Set_Reading_Area) {
						CmdList[CmdIndex] = CmdList[J];
						CmdList[J] = C_None;
						CmdIndex++;
					}
				}
				break;
			case 's' :
				CmdList[CmdIndex] =C_Start_Scanning;
				CmdIndex++;
				if (argc > 2) {
					FileName = argv[2];
					if ((FileName[0]=='-')  && (FileName[1]=='\0')) FileName = "stdout";
					I = 255; argc--;  argv++;
				}
				break;	
			case 'v' :
				Debug++;
				break;
			case 'x' :
				Quiet++;
				break;
			case 'y' :
				if (argc > 2) {
					Protocol = argv[2];
					I = 255; argc--;  argv++;
				}
				break;
			case 'z' :
				CmdList[CmdIndex] =C_Set_Zoom;
				CmdIndex++;
				if (argc > 2) {
					Zoom1 = argv[2]; 
					Zoom2 = argv[2];
					I = 255; argc--;  argv++;
				}
				/* Check all parms before this. If any -A command is found,
				 	force the -A to come AFTER this command, as this command
					causes a reset of the area to its maximum value. */
				K = CmdIndex;			/* Only look so far */
				for(J=0; J<K; J++) {
					if (CmdList[J]==C_Set_Reading_Area) {
						CmdList[CmdIndex] = CmdList[J];
						CmdList[J] = C_None;
						CmdIndex++;
					}
				}
				break;

			/* These options correspond directly to scanner commands */
			case 'I' :
				CmdList[CmdIndex] = C_Request_Identity;
				CmdIndex++;
				break;
			case 'F' :
				CmdList[CmdIndex] =C_Request_Status;
				CmdIndex++;
				break;
			case 'q' :
			case 'S' :
				CmdList[CmdIndex] =C_Request_Condition;
				CmdIndex++;
				break;
			case 'G' :
				CmdList[CmdIndex] =C_Start_Scanning;
				CmdIndex++;
				break;
			case 'C' :
				CmdList[CmdIndex] =C_Set_Color_Mode;
				CmdIndex++;
				if (argc > 2) {
					ColorMode = argv[2];
					I = 255; argc--;  argv++;
				}
				break;
			case 'D' :
				CmdList[CmdIndex] =C_Set_Data_Format;
				CmdIndex++;
				if (argc > 2) {
					DataFormat = argv[2];
					I = 255; argc--;  argv++;
				}
				break;
			case 'R' :
				CmdList[CmdIndex] =C_Set_Resolution;
				CmdIndex++;
				if (argc > 3) {
					Resolution1 = argv[2]; argc--;  argv++;
					Resolution2 = argv[2];
					I = 255; argc--;  argv++;
				}
				/* Check all parms before this. If any -A command is found,
				 	force the -A to come AFTER this command, as this command
					causes a reset of the area to its maximum value. */
				K = CmdIndex;			/* Only look so far */
				for(J=0; J<K; J++) {
					if (CmdList[J]==C_Set_Reading_Area) {
						CmdList[CmdIndex] = CmdList[J];
						CmdList[J] = C_None;
						CmdIndex++;
					}
				}
				break;
			case 'H' :
				CmdList[CmdIndex] =C_Set_Zoom;
				CmdIndex++;
				if (argc > 3) {
					Zoom1 = argv[2]; argc--;  argv++;
					Zoom2 = argv[2];
					I = 255; argc--;  argv++;
				}
				/* Check all parms before this. If any -A command is found,
				 	force the -A to come AFTER this command, as this command
					causes a reset of the area to its maximum value. */
				K = CmdIndex;			/* Only look so far */
				for(J=0; J<K; J++) {
					if (CmdList[J]==C_Set_Reading_Area) {
						CmdList[CmdIndex] = CmdList[J];
						CmdList[J] = C_None;
						CmdIndex++;
					}
				}
				break;
			case 'a' :
			case 'A' :
				CmdList[CmdIndex] =C_Set_Reading_Area;
				CmdIndex++;
				if (argc > 5) {
					ReadingArea1 = argv[2]; argc--;  argv++;
					ReadingArea2 = argv[2]; argc--;  argv++;
					ReadingArea3 = argv[2]; argc--;  argv++;
					ReadingArea4 = argv[2];
					I = 255; argc--;  argv++;
				}
				break;
			case 'L' :
				CmdList[CmdIndex] =C_Set_Brightness;
				CmdIndex++;
				if (argc > 2) {
					Brightness = argv[2];
					I = 255; argc--;  argv++;
				}
				break;
			case 'Z' :
				CmdList[CmdIndex] =C_Set_Gamma_Correction;
				CmdIndex++;
				if (argc > 2) {
					GammaCorrection = argv[2];
					I = 255; argc--;  argv++;
				}
				break;
			case 'B' :
				CmdList[CmdIndex] =C_Set_Halftoning_Mode;
				CmdIndex++;
				if (argc > 2) {
					HalfToningMode = argv[2];
					I = 255; argc--;  argv++;
				}
				break;
			case 'M' :
				CmdList[CmdIndex] =C_Set_Color_Correction;
				CmdIndex++;
				if (argc > 2) {
					ColorCorrection = argv[2];
					I = 255; argc--;  argv++;
				}
				break;
			case '@' :
				CmdList[CmdIndex] =C_Initialize;
				CmdIndex++;
				break;
			case 'h' :
				Usage();
				return(1);
			default :
				fprintf(stderr,"%s: Unknown option '%c'\n\n", MeName, argv[1][I] );
				Usage();
				return(1);
			}
		}
	argc--;  argv++;
 	}
	CmdList[CmdIndex] =C_End;		/* Mark end of list of commands */
	
	/*	There should be no parms left - Everything is specified as an option!  */
	if (argc > 1)  {
		fprintf(stderr,"%s: Unknown parameter '%s'\nTry '%s -h' for help.\n",
			MeName, argv[1], MeName);
		return(1);
	}
	
	/* Open the scanner. Set some global data so we can read/write to scanner */
	if (FindScanner(ScannerDevice, Protocol, Target, LineSpeed)<0) return(3);

	/*	Now do whatever it is that was requested, in the order it was requested. */
	for(CmdIndex=0; CmdList[CmdIndex]!=C_End; CmdIndex++) {
		Cmd = CmdList[CmdIndex];
		switch (Cmd) {
			case C_Request_Identity :
				if (CmdReqData(Cmd, &ScannerIdentifyDB.header)) {
					/* Check format of data block */
					printf("Scanner id= %2.2s Resolutions= ", 
						ScannerIdentifyDB.identification);
					Res = (char *) &(ScannerIdentifyDB.res);
					while(*Res=='R') {
						printf("%d ", ByteCounter(Res+1));
						Res = Res + 3;
					};
					/* Reading area should follow. We should be at an 'A' */
					if (*Res == 'A') {
						printf("Reading area= %dx%d\n",
						ByteCounter(Res+1), ByteCounter(Res+3));
				}
					else {
						if (!Quiet)
							fprintf(stderr,"%s: Identify block format error: Reading Area\n",
							MeName);
						Error++;
					}
				}
				break;
			case C_Request_Status:
				if (CmdReqData(Cmd, &ScannerStatusDB.header)) {
					fprintf(stderr, "Status= %02X\n", ScannerStatusDB.header.status);
				}
				break;
			case C_Request_Condition :
				if (CmdReqData(Cmd, &ScannerConditionDB.header)) {
					if (Debug) {
						if (ScannerConditionDB. C != 'C')
							fprintf(stderr,"Condition block format error: Color_Mode\n");
						if (ScannerConditionDB. R != 'R')
							fprintf(stderr,"Condition block format error: Output  Resolution\n");
						if (ScannerConditionDB. A != 'A')
							fprintf(stderr,"Condition block format error: Reading Area\n");
						if (ScannerConditionDB. D != 'D')
							fprintf(stderr,"Condition block format error: Format Data\n");
						if (ScannerConditionDB. B != 'B')
							fprintf(stderr,"Condition block format error: Halftoning Mode\n");
						if (ScannerConditionDB. L != 'L')
							fprintf(stderr,"Condition block format error: Brightness\n");
						if (ScannerConditionDB. Z != 'Z')
							fprintf(stderr,"Condition block format error: Gamma Correction\n");
						if (ScannerConditionDB. H != 'H')
							fprintf(stderr,"Condition block format error: Zoom\n");
						if (ScannerConditionDB. M != 'M')
							fprintf(stderr,"Condition block format error: Color Correction\n");
					}
					fprintf(stderr, "Color Mode= %02X ", ScannerConditionDB.Color_Mode);
					fprintf(stderr, "Output Resolution= %d,%d ",
						ByteCounter(&ScannerConditionDB.outputmainscanresolution),
						ByteCounter(&ScannerConditionDB.outputsubscanresolution));
					fprintf(stderr, "Reading Area offset= %d,%d ",
						ByteCounter(&ScannerConditionDB.readingareamainscanoffset),
						ByteCounter(&ScannerConditionDB.readingareasubscanoffset));
					fprintf(stderr, "Reading Area= %dx%d ",
						ByteCounter(&ScannerConditionDB.readingareamainscanlength),
						ByteCounter(&ScannerConditionDB.readingareasubscanlength));
					fprintf(stderr, "Data Format= %02X ", ScannerConditionDB.Format_Data);
					fprintf(stderr, "Halftoning Mode= %02X ", 
						ScannerConditionDB.HaltToningMode);
					fprintf(stderr, "Brightness= %02X ", ScannerConditionDB.Brightness);
					fprintf(stderr, "Gamma Correction= %02X ",
						ScannerConditionDB.Gamma_Correction);
					fprintf(stderr, "Zoom area= %d,%d ",
						(int) ScannerConditionDB.zoommainscan,
						(int) ScannerConditionDB.zoomsubscan);
					fprintf(stderr, "Color Correction= %02X\n",
						ScannerConditionDB.ColorCorrection);
				}
				break;
			case C_Start_Scanning :
				if (ScannerIdentifyDB.identification[1] == ' ')    /* Get model of scanner */
					CmdReqData(C_Request_Identity, &ScannerStatusDB.header);
				if (!DoScan(FileName, DocName)) Error++;
				break;
			case C_Set_Color_Mode:
				if (ColorMode == NULL) {	/* Set color based on model */
					if (ScannerIdentifyDB.identification[1] == ' ') 						CmdReqData(C_Request_Identity, &ScannerIdentifyDB.header);
					if (ScannerIdentifyDB.identification[1] != '3') ColorMode = "2";
					else ColorMode = "1";
				}
				sscanf(ColorMode,"%x", &HexInt);
				HexByte = (u_char) 255 & HexInt;
				if (!CmdWithParm(Cmd, &HexByte , 1)) Error++;
				break;
			case C_Set_Data_Format :
				sscanf(DataFormat,"%d", &HexInt);
				HexByte = (u_char) 255 & HexInt;
				if (!CmdWithParm(Cmd, &HexByte , 1)) Error++;
				break;
			case C_Set_Resolution:
				/* First get the resolutions supported by the scanner. Make resolutions
					machine normal numbers */
				if (!CmdReqData(C_Request_Identity, &ScannerIdentifyDB.header)) {
					if (!Quiet) fprintf(stderr,"%s: Identify block format error\n", MeName);
					Error++;
					break;
				} 
				Res = (char *) &(ScannerIdentifyDB.res);
				while(*Res=='R')  {
					shortI = ByteCounter(Res+1);
					bcopy(&shortI, Res+1, 2);
					Res = Res + 3;
				};
				/* Now convert resolutions specified by the user and choose nearest resolution
					this scanner supports. */
				sscanf(Resolution1,"%d", &HexInt);
				Res = (char *) &(ScannerIdentifyDB.res);
				while(*Res=='R') {
					bcopy(Res+1, &shortI, 2);
					if (HexInt <= shortI) break;
					Res = Res + 3;
				};
				HexInt = shortI;
				MakeByteCounter(HexInt, &HexParm.HexShort);
				sscanf(Resolution2,"%d", &HexInt);
				Res = (char *) &(ScannerIdentifyDB.res);
				while(*Res=='R') {
					bcopy(Res+1, &shortI, 2);
					if (HexInt <= shortI) break;
					Res = Res + 3;
				};
				HexInt = shortI;
				MakeByteCounter(HexInt, &HexParm.HexShort1);
				if (!CmdWithParm(Cmd, &HexParm , 4)) Error++;
				break;
			case C_Set_Zoom :
				sscanf(Zoom1,"%d", &HexInt);
				if (HexInt < 50) HexInt = 50;
				if (HexInt > 200) HexInt = 200;
				HexBytes[0] = (u_char) 255 & HexInt;
				sscanf(Zoom2,"%d", &HexInt);
				if (HexInt < 50) HexInt = 50;
				if (HexInt > 200) HexInt = 200;
				HexBytes[1] = (u_char) (255 & HexInt);
				if (!CmdWithParm(Cmd, &HexBytes , 2)) Error++;
				break;
			case C_Set_Reading_Area:
				sscanf(ReadingArea1,"%d", &HexInt);
				I = HexInt;
				MakeByteCounter(HexInt, &HexParm.HexShort);
				sscanf(ReadingArea2,"%d", &HexInt);
				MakeByteCounter(HexInt, &HexParm.HexShort1);
				sscanf(ReadingArea3,"%d", &HexInt);
				HexInt = HexInt&0xfff8;	/* Force to multiple of 8 */
				if (HexInt + I > 4096) HexInt = 4096 - I;
				MakeByteCounter(HexInt, &HexParm.HexShort2);
				sscanf(ReadingArea4,"%d", &HexInt);
				MakeByteCounter(HexInt, &HexParm.HexShort3);
				if (!CmdWithParm(Cmd, &HexParm , 8)) Error++;
				break;
			case C_Set_Brightness:
				sscanf(Brightness,"%x", &HexInt);
				HexByte = (u_char) 255 & HexInt;
				if (!CmdWithParm(Cmd, &HexByte , 1)) Error++;
				break;
			case C_Set_Gamma_Correction:
				sscanf(GammaCorrection,"%x", &HexInt);
				HexByte = (u_char) 255 & HexInt;
				if (!CmdWithParm(Cmd, &HexByte , 1)) Error++;
				break;
			case C_Set_Halftoning_Mode :
				sscanf(HalfToningMode,"%x", &HexInt);
				HexByte = (u_char) 255 & HexInt;
				if (!CmdWithParm(Cmd, &HexByte , 1)) Error++;
				break;
			case C_Set_Color_Correction :
				sscanf(ColorCorrection,"%x", &HexInt);
				HexByte = (u_char) 255 & HexInt;
				if (!CmdWithParm(Cmd, &HexByte , 1)) Error++;
				break;
			case C_Initialize :
				if (CmdNoParms(Cmd)) Error++;
				break;
			case C_None :
				break;
			default :
				fprintf(stderr,"usage: %s [-options] [file]\n", MeName);
				return(1);
			}
		}
	return(Error);			/* Exit  0 == All OK,  >0 == # errors we encountered */
}

/*
Usage ()

Function:		Hint  to user how to use this pgm

Returns:		Never
*/
void Usage()
{
	fprintf(stderr, "Usage: %s -options\n\tScans data from Epson 300C scanner creating TIFF or PPM output files.\n\tVersion %s\n\nOptions:\n", MeName, VERSION);
	fprintf(stderr,"\ta P Q X Y    Set to scan area at offset P,Q and X by Y pixels\n");
	fprintf(stderr,"\tb N\tSerial line speed: 9600 or 19200\n");
	fprintf(stderr,"\tc\tSet to color 24 bits/pixel\n");
	fprintf(stderr,"\td dev\tScanner device name (%s)\n", DEVICE);
	fprintf(stderr,"\tf Addr\tSCSI device address. Default=search for device.\n");
	fprintf(stderr,"\tg\tSet to gray-scale 8 bits/pixel\n");
	fprintf(stderr,"\tg4\tSet to gray-scale 4 bits/pixel\n");
	fprintf(stderr,"\tg2\tSet to gray-scale 2 bits/pixel\n");
	fprintf(stderr,"\ti\tInitialize scanner\n");
	fprintf(stderr,"\tm\tSet to monochrome  1 bit/pixel\n");
	fprintf(stderr,"\tn name\tName of document\n");
	fprintf(stderr,"\tq\tQuery status of scanner\n");
	fprintf(stderr,"\tr N\tSet resolution to N dpi (50-600)\n");
	fprintf(stderr,"\ts File\tScan at current settings to File.\n");
 	fprintf(stderr,"\tv\tVerbose level\n");
	fprintf(stderr,"\tx\tQuiet mode\n");
	fprintf(stderr,"\ty Protocol\tForce device protocol to serial, scsi, or parallel\n");
 	fprintf(stderr,"\tz N\tSet zoom level to N%%\n");
	fprintf(stderr,"\nFor example: %s -a 200 400 350 450 -g -s /tmp/test.tiff\n", MeName);
	exit(1);
}


