/*	Scan routines for Epson 300C color scanner

Copyright (c) Terence P. Gliedt, 1992.
	This program may be freely redistributable in its original, unmodified form.
	The author grants you a nonexclusive, royalty-free license to use this software,
	and modify it as needed provided that this copyright notice remains, and any
	changes are clearly indicated as such.  No part may be incorporated in products or
	packages sold for profit without express permission of the copyright holder.
	The author assumes no responsibility for the consequences of using this program
	and no warranty of any sort is expressed or implied.
*/
#include <stdio.h>
#include <sys/file.h>
#include <sys/time.h>
#include <signal.h>
#include "es300c.h"
#include "tiffio.h"		/* From TIFF Software "Release" 3.00 */

int ScanRecover();

/*
ScanRecover (Intrpt)

Function:		Signal handler for errors while a scan is in progress

Returns:		0 always
*/
int ScanRecover()
{

	/* We come here most likely because the user has hit cntrl-C during a scan.
		But it doesn't much matter - for whatever reason we end up here, our
		only task is to stop the scan gracefully. Set a flag so DoScan will
		terminate nicely. */
	Abort++;
	return(0);
}

/*
DoColor24Page(ScannerConditionDB, FileN, DocN)

Function:		Scan 24 bit color data from scanner, writing results to file FileN.
			Set name of document to DocN. File created is a ppm format stream.
			To help figure out state of scanner, ScannerConditionDB is the
			resultant structure from a C_Request_Condition request.
			This only suppoorts the lower end model - B2 (GT-1000).

Returns:		Number of bytes read. 0 means the scanfailed
*/
int DoColor24Page(struct Condition_Data_Block *ScannerConditionDB, char *FileN, char *DocN)
{
	char CmdToIssue[2];	/* Command to send to scanner */
	struct IDB Header;		/* Header data preceding each chunk of data */
	char PPMBuffer[30];	/* ASCII formatted PPM data. R G B */
	u_char *DataBufferR;	/* Read red image data into here */
	u_char *DataBufferG;	/* Read green image data into here */
	u_char *DataBufferB;	/* Read blue image data into here */
	u_char *Buffer;			/* Ptr to DataBuffer? */
	int  DataSize;			/* Number of bytes of data to come from the scanner */
	int  DataWidth;			/* Number of bits in each scanline */
	int  DataRows;			/* Number of rows of scanlines to read/write */
	int  DataBPP;			/* Number bits per pixel */
	int BytesPerRow;		/* Number of bytes per row */
	int  DataRes;			/* Resolution of this picture */
	int  TotalLen;			/* Number of bytes read */
	int  Progress = 0;		/* Every N bytes, show that something is happening */
	int  Len;				/* Number of bytes to read */
	int Row;				/* To figure count what scanline we are reading */
	long T;				/* For messing with the time */
	int II;
	struct tm *Time;		/* What we get back from localtime */
	FILE *File;			/* File to be created */
	
	/* Open the File if we need to so we can write the scan data to it. */
	if (strcmp(FileN, "stdout")) {
	    	if((File = fopen(FileN, "w")) == NULL) {
			fprintf(stderr, "%s: Unable to open file '%s'\n", MeName, FileN);
			return(0);
		}
	}
	else {
		File = stdout;
	}

	/* This is a scan of a color image */
	DataBPP = 24;
	DataWidth = ByteCounter(&(ScannerConditionDB->readingareamainscanlength));
	DataRows = ByteCounter(&(ScannerConditionDB->readingareasubscanlength));
	DataSize = DataRows * DataWidth * DataBPP/ 8;
	BytesPerRow = DataSize / (DataRows*3);
	DataRes = ByteCounter(&ScannerConditionDB->outputmainscanresolution);

	/* Get some big areas for the raw red & green data. Once we have this, the blue data can
		be read in smaller chunks. */
	II = (DataSize/3)+8;
	DataBufferR = (u_char *) malloc(II);
	DataBufferG = (u_char *) malloc(II);
	DataBufferB = (u_char *) malloc(II);

	/* Now start building the RAWBITS PPM file */
	sprintf(DataBufferB, "P6\n%d %d\n255\n# Scanned on Epson 300C\n",
		 DataWidth, DataRows);
	fwrite(DataBufferB, strlen(DataBufferB), 1, File);   /* Magic #, Width Height  + max colors */
	if (DocN)  {
		fwrite("# ", 2, 1, File);
		fwrite(DocN, strlen(DocN), 1, File);   /* Document Name */
		fwrite("\n", 1, 1, File);
	}
	T = time(0);
	Time = localtime(&T);				/* Get nice structure */
	sprintf(DataBufferB, "# %4d:%02d:%02d %02d:%02d:%02d\n",
		Time->tm_year+1900, Time->tm_mon+1, Time->tm_mday,
		Time->tm_hour, Time->tm_min, Time->tm_sec);
	fwrite(DataBufferB, strlen(DataBufferB), 1, File);   /* DATETIME */

	TotalLen = 0;				/* Total bytes written to file */
	if (!Quiet) 
		fprintf(stderr,"Creating %d dpi color image %dx%dx%d at %d%% ",
			DataRes, DataWidth, DataRows, DataBPP, 
			(int) ScannerConditionDB->zoommainscan);

	/* Set a signal handler. About the only thing to worry about is the user hitting Cntrl-C.
		If this happens the scanner will enter an E-1 state (interface error). Only
		a power-on-reset will reset the scanner. Try our best to avoid this. */
	signal(SIGINT, ScanRecover);
	if (ScannerType == SERIAL)  signal(SIGHUP, ScanRecover);

	/* Build the stream to be sent to the device */
	CmdToIssue[0] = ESC;
	CmdToIssue[1] = (char) C_Start_Scanning;
        DEBUG("\tSending cmd '%s' to scanner\n", Hex(CmdToIssue, 2));
	if (!DoWrite((char *) &CmdToIssue, 2)) return(0);

	/* Read red, green and blue scan pages in. These cannot be written out until green and red
		are read in. Then as the blue is read, the red can be written out. */
	for(TotalLen=0,  II=0; II<3; II++) {
		Buffer = DataBufferG;
		if (II==1) Buffer = DataBufferR;
		if (II==2) Buffer = DataBufferB;
		for (Row=0; Row <DataRows; Row++) {
			/* Now read a data block. Data is preceded by IDB, then real data */
			if (!DoRead ( (char *) &Header,  sizeof(struct IDB))) return(0);
			if (Debug) {
				if (Header.header == STX) fprintf(stderr,"\tInformation block=%s\n",
					Hex((char *)&Header, sizeof(struct IDB)));
			}

			/* Data from scanner better be the expected size or we are really messed up */
			Len = ByteCounter(&Header.length);
			if (Len != BytesPerRow) {
				fprintf(stderr, "%s: Scanner delivered %d bytes when expecting %d.\n",
					MeName, Len, BytesPerRow);
				Abort++;
			}

			/* R E A D the data into the proper color Buffer */
			if (!DoRead (Buffer,  Len)) return(0);
			if (Debug) {
				if (Debug>1)  fprintf(stderr,"\tReceived %d bytes=%s\n", Len, Hex(Buffer, Len));
				else fprintf(stderr,"\tReceived %d bytes\n", Len);
			}
		
			/* Cancel by user or internal error - both mean stop the presses! */
			if (Abort) {
				fprintf(stderr, "%s: Scan cancelled. %d bytes of image data written to '%s'\n",
					MeName, TotalLen, FileN);
				CmdToIssue[0] = CAN;		/* Cancel the scan */
			        DEBUG("\tSending CAN '%s' to scanner\n", Hex(CmdToIssue, 1));
				DoWrite((char *) &CmdToIssue, 1);
				DoRead(&CmdToIssue, 1) ; 	 /* Get ACK to CAN */
				(void) fclose(File);
				return(0);
			}

			/* Got data, now send ACK telling scanner we got it, but only more data is to come */
			Buffer = Buffer + Len;		/* Next data goes here */
			TotalLen = TotalLen + Len;	/* Keep track of # bytes read */
			Progress = Progress + Len;	/* Keep track of # bytes read */
			if (!(Header.status & Status_AreaEnd)) {
				CmdToIssue[0] = ACK;
			        DEBUG("\tSending ACK '%s' to scanner\n", Hex(CmdToIssue, 1));
				if (!DoWrite((char *) &CmdToIssue, 1)) return(0);
			}

			/* If not debugging, show some progress with a row of dots */
			if ((!Debug) && (Progress>=100000)) {
				if (!Quiet) fprintf(stderr,".");
				Progress = 0;
			}
		}
	}
	/* Now we have data in the R, G & B buffers. Write this out .
		For each byte in each row, generate a simple ASCII string of G B R.
		I figure we are gonna quit real soon, so I will not bother freeing the malloc buffers. */
	for (II=0; II <TotalLen; II=II+3)  {
		PPMBuffer[0] = *DataBufferG++;
		PPMBuffer[1] = *DataBufferB++;
		PPMBuffer[2] = *DataBufferR++;
		fwrite(PPMBuffer, 3, 1, File);   /* Green Blue Red value */
	} 

	if (!Quiet) fprintf(stderr,"  %d bytes of PPM image data written to file '%s'\n", TotalLen, FileN);
	(void) fclose(File);
	return(TotalLen);
}

/*
DoColor24Scan(ScannerConditionDB, FileN, DocN)

Function:		Scan 24 bit color data from scanner, writing results to file FileN.
			Set name of document to DocN. File created is a ppm format stream.
			To help figure out state of scanner, ScannerConditionDB is the
			resultant structure from a C_Request_Condition request.
			This only suppoorts the top end model - B3 (GT-4000 and GT-6000).

Returns:		Number of bytes read. 0 means the scanfailed
*/
int DoColor24Scan(struct Condition_Data_Block *ScannerConditionDB, char *FileN, char *DocN)
{
	char CmdToIssue[2];	/* Command to send to scanner */
	struct IDB Header;		/* Header data preceding each chunk of data */
	char PPMBuffer[30];	/* ASCII formatted PPM data. R G B */
	u_char DataBufferR[4096];		/* Read red image data into here */
	u_char DataBufferG[4096];		/* Read green image data into here */
	u_char DataBufferB[4096];		/* Read blue image data into here */
	u_char *Buffer;			/* Ptr to DataBuffer? */
	int  DataSize;			/* Number of bytes of data to come from the scanner */
	int  DataWidth;			/* Number of bits in each scanline */
	int  DataRows;			/* Number of rows of scanlines to read/write */
	int  DataBPP;			/* Number bits per pixel */
	int BytesPerRow;		/* Number of bytes per row */
	int  DataRes;			/* Resolution of this picture */
	int  TotalLen;			/* Number of bytes read */
	int  Progress = 0;		/* Every N bytes, show that something is happening */
	int  Len;				/* Number of bytes to read */
	int Row;				/* To figure count what scanline we are reading */
	long T;				/* For messing with the time */
	int II;
	struct tm *Time;		/* What we get back from localtime */
	FILE *File;			/* File to be created */
	
	/* Open the File if we need to so we can write the scan data to it. */
	if (strcmp(FileN, "stdout")) {
	    	if((File = fopen(FileN, "w")) == NULL) {
			fprintf(stderr, "%s: Unable to open file '%s'\n", MeName, FileN);
			return(0);
		}
	}
	else {
		File = stdout;
	}

	/* This is a scan of a color image */
	DataBPP = 24;
	DataWidth = ByteCounter(&(ScannerConditionDB->readingareamainscanlength));
	DataRows = ByteCounter(&(ScannerConditionDB->readingareasubscanlength));
	DataSize = DataRows * DataWidth * DataBPP/ 8;
	BytesPerRow = DataSize / (DataRows*3);
	DataRes = ByteCounter(&ScannerConditionDB->outputmainscanresolution);

	/* Now start building the RAWBITS PPM file */
	sprintf(DataBufferR, "P6\n%d %d\n255\n# Scanned on Epson 300C\n",
		 DataWidth, DataRows);
	fwrite(DataBufferR, strlen(DataBufferR), 1, File);   /* Magic #, Width Height  + max colors*/
	if (DocN)  {
		fwrite("# ", 2, 1, File);
		fwrite(DocN, strlen(DocN), 1, File);   /* Document Name */
		fwrite("\n", 1, 1, File);
	}
	T = time(0);
	Time = localtime(&T);				/* Get nice structure */
	sprintf(&DataBufferR, "# %4d:%02d:%02d %02d:%02d:%02d\n",
		Time->tm_year+1900, Time->tm_mon+1, Time->tm_mday,
		Time->tm_hour, Time->tm_min, Time->tm_sec);
	fwrite(DataBufferR, strlen(DataBufferR), 1, File);   /* DATETIME */
	
	TotalLen = 0;				/* Total bytes written to file */
	if (!Quiet) 
		fprintf(stderr,"Creating %d dpi color image %dx%dx%d at %d%% ",
			DataRes, DataWidth, DataRows, DataBPP, 
			(int) ScannerConditionDB->zoommainscan);

	/* Set a signal handler. About the only thing to worry about is the user hitting Cntrl-C.
		If this happens the scanner will enter an E-1 state (interface error). Only
		a power-on-reset will reset the scanner. Try our best to avoid this. */
	signal(SIGINT, ScanRecover);
	if (ScannerType == SERIAL)  signal(SIGHUP, ScanRecover);

	/* Build the stream to be sent to the device */
	CmdToIssue[0] = ESC;
	CmdToIssue[1] = (char) C_Start_Scanning;
        DEBUG("\tSending cmd '%s' to scanner\n", Hex(CmdToIssue, 2));
	if (!DoWrite((char *) &CmdToIssue, 2)) return(0);

	/* Read red, green and blue scanlines in. Then write this scan line out  as sets of triples. */
	for(TotalLen=0,  Row=0; Row <DataRows; Row++) {
		for (II=0; II<3; II++) {
			Buffer = &DataBufferG[0];
			if (II==1) Buffer = &DataBufferR[0];
			if (II==2) Buffer = &DataBufferB[0];
			/* Now read a data block. Data is preceded by IDB, then real data */
			if (!DoRead ( (char *) &Header,  sizeof(struct IDB))) return(0);
			if (Debug) {
				if (Header.header == STX) fprintf(stderr,"\tInformation block=%s\n",
					Hex((char *)&Header, sizeof(struct IDB)));
			}

			/* Data from scanner better be the expected size or we are really messed up */
			Len = ByteCounter(&Header.length);
			if (Len != BytesPerRow) {
				fprintf(stderr, "%s: Scanner delivered %d bytes when expecting %d.\n",
					MeName, Len, BytesPerRow);
				Abort++;
			}

			/* R E A D the data into DataBuffer? - this happens three buffers at a time. */
			if (!DoRead (Buffer,  Len)) return(0);
			if (Debug) {
				if (Debug>1)  fprintf(stderr,"\tReceived %d bytes=%s\n", Len, Hex(Buffer, Len));
				else fprintf(stderr,"\tReceived %d bytes\n", Len);
			}
		
			/* Cancel by user or internal error - both mean stop the presses! */
			if (Abort) {
				fprintf(stderr, "%s: Scan cancelled. %d bytes of image data written to '%s'\n",
					MeName, TotalLen, FileN);
				CmdToIssue[0] = CAN;		/* Cancel the scan */
			        DEBUG("\tSending CAN '%s' to scanner\n", Hex(CmdToIssue, 1));
				DoWrite((char *) &CmdToIssue, 1);
				DoRead(&CmdToIssue, 1) ; 	 /* Get ACK to CAN */
				(void) fclose(File);
				return(0);
			}

			/* Got data, now send ACK telling scanner we got it, but only more data is to come */			TotalLen = TotalLen + BytesPerRow;   /* Keep track of # bytes read */
			Progress = Progress + BytesPerRow;  /* Keep track of # bytes read */
			if (!(Header.status & Status_AreaEnd)) {
				CmdToIssue[0] = ACK;
			        DEBUG("\tSending ACK '%s' to scanner\n", Hex(CmdToIssue, 1));
				if (!DoWrite((char *) &CmdToIssue, 1)) return(0);
			}

			/* If not debugging, show some progress with a row of dots */
			if ((!Debug) && (Progress>=100000)) {
				if (!Quiet) fprintf(stderr,".");
				Progress = 0;
			}
		}
		
		/* Now we have data in the R, G & B buffers. Write this out .
		 For each byte in each row, generate a simple ASCII string of G B R */
		for (II=0; II <BytesPerRow; II++)  {
			PPMBuffer[0] = DataBufferG[II];
			PPMBuffer[1] = DataBufferB[II];
			PPMBuffer[2] = DataBufferR[II];
			fwrite(PPMBuffer, 3, 1, File);   /* Green Blue Red value */
		} 
	}

	if (!Quiet) fprintf(stderr,"  %d bytes of PPM image data written to file '%s'\n", TotalLen, FileN);
	(void) fclose(File);
	return(TotalLen);
}

/*
DoGrayScan(ScannerConditionDB,  FileN, DocN)

Function:		Scan non-color data from scanner, writing results to file FileN. Set name of
			document to DocN. File created is a TIFF file.
			To help figure out state of scanner, ScannerConditionDB is the
			resultant structure from a C_Request_Condition request.

Returns:		Number of bytes read. 0 means the scanfailed
*/
int DoGrayScan(struct Condition_Data_Block *ScannerConditionDB, char *FileN, char *DocN)
{
	char CmdToIssue[2];	/* Command to send to scanner */
	struct IDB Header;		/* Header data preceding each chunk of data */
	u_char DataBuffer[4096];		/* Read image data into here */
	int  DataSize;			/* Number of bytes of data to come from the scanner */
	u_long  DataWidth;		/* Number of bits in each scanline */
	u_long  DataRows;		/* Number of rows of scanlines to read/write */
	int  DataBPP;			/* Number bits per pixel */
	int BytesPerRow;		/* Number of bytes per row */
	int  DataRes;			/* Resolution of this picture */
	int  TotalLen;			/* Number of bytes read */
	int  Progress = 0;		/* Every N bytes, show that something is happening */
	int  Len;				/* Number of bytes to read */
	unsigned Row;			/* To figure count what scanline we are reading */
	u_long RowsPerStrip;	/* This seems to be important for some reason */
	char *P;
	long T;				/* For messing with the time */
	struct tm *Time;		/* What we get back from localtime */
	TIFF *File;			/* TIFF file to create handle */

	/* Open the File if we need to so we can write the scan data to it. */
	if (strcmp(FileN, "stdout")) {
	    	if((File = TIFFOpen(FileN, "w")) == NULL) {
			fprintf(stderr, "%s: Unable to open file '%s'\n", MeName, FileN);
			return(0);
		}
	}
	else {
		fprintf(stderr, "%s: TIFF output may not be written to stdout.\n", MeName);
		return(0);
	}

	DataBPP = ScannerConditionDB->Format_Data;
	if (ScannerConditionDB->Format_Data > 1) P = "gray scale";
	else P = "monochrome";
	DataWidth = ByteCounter(&(ScannerConditionDB->readingareamainscanlength));
	DataRows = ByteCounter(&(ScannerConditionDB->readingareasubscanlength));
	DataSize = DataRows * DataWidth * DataBPP/ 8;
	BytesPerRow = DataSize / DataRows;
		
	/* Now start building the TIFF header */
	TIFFSetField(File, TIFFTAG_IMAGEWIDTH, (u_long) DataWidth);
	TIFFSetField(File, TIFFTAG_IMAGELENGTH, (u_long) DataRows);
	TIFFSetField(File, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
	TIFFSetField(File, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
	RowsPerStrip = (8*1024)/DataWidth;
	TIFFSetField(File, TIFFTAG_SAMPLESPERPIXEL, (u_short) 1);
	TIFFSetField(File, TIFFTAG_BITSPERSAMPLE,
		(u_short) ScannerConditionDB->Format_Data);
	TIFFSetField(File, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
	TIFFSetField(File, TIFFTAG_COMPRESSION, COMPRESSION_LZW);
	TIFFSetField(File, TIFFTAG_ROWSPERSTRIP, RowsPerStrip);
	TIFFSetField( File, TIFFTAG_IMAGEDESCRIPTION, "Scanned on Epson 300C" );
	if (DocN)  TIFFSetField(File, TIFFTAG_DOCUMENTNAME, DocN);

	T = time(0);
	Time = localtime(&T);				/* Get nice structure */
	sprintf(&DataBuffer, "%4d:%02d:%02d %02d:%02d:%02d",
		(int) (Time->tm_year)+1900, (int) (Time->tm_mon)+1, (int) Time->tm_mday,
		(int) Time->tm_hour, (int) Time->tm_min, (int) Time->tm_sec);
	TIFFSetField(File, TIFFTAG_DATETIME,  &DataBuffer);
	TIFFSetField(File, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
	DataRes = ByteCounter(&(ScannerConditionDB->outputmainscanresolution));
	TIFFSetField(File, TIFFTAG_XRESOLUTION,
		(float) ByteCounter(&(ScannerConditionDB->outputmainscanresolution)));
	TIFFSetField(File, TIFFTAG_YRESOLUTION,
		(float) ByteCounter(&(ScannerConditionDB->outputsubscanresolution)));

	TotalLen = 0;				/* Total bytes written to file */
	if (!Quiet) 
		fprintf(stderr,"Creating %d dpi %s image %dx%dx%d at %d%% ",
			DataRes, P, DataWidth, DataRows, DataBPP, 
			(int) ScannerConditionDB->zoommainscan);

	/* Set a signal handler. About the only thing to worry about is the user hitting Cntrl-C.
		If this happens the scanner will enter an E-1 state (interface error). Only
		a power-on-reset will reset the scanner. Try our best to avoid this. */
	signal(SIGINT, ScanRecover);
	if (ScannerType == SERIAL)  signal(SIGHUP, ScanRecover);

	/* Build the stream to be sent to the device */
	CmdToIssue[0] = ESC;
	CmdToIssue[1] = (char) C_Start_Scanning;
        DEBUG("\tSending cmd '%s' to scanner\n", Hex(CmdToIssue, 2));
	if (!DoWrite((char *) &CmdToIssue, 2)) return(0);

	/* Read image a scanline at a time, then dump to file.  */
	for(TotalLen=0, Row=0; Row < DataRows; Row++) {
		/* Now read a data block. Data is preceded by IDB, then real data */
		if (!DoRead ( (char *) &Header,  sizeof(struct IDB))) return(0);
		if (Debug) {
			if (Header.header == STX) fprintf(stderr,"\tInformation block=%s\n",
				Hex((char *)&Header, sizeof(struct IDB)));
		}

		/* Data from scanner better be the expected size or we are really messed up */
		Len = ByteCounter(&Header.length);
		if (Len > BytesPerRow) {
			fprintf(stderr, "%s: Scanner delivered %d bytes when expecting %d.\n",
				MeName, Len, BytesPerRow);
			Abort++;
		}

		/* R E A D the data in */
		if (!DoRead (DataBuffer,  Len)) return(0);
		if (Debug) {
			if (Debug>1)  fprintf(stderr,"\tReceived %d bytes=%s\n", Len, Hex(DataBuffer, Len));
			else fprintf(stderr,"\tReceived %d bytes\n", Len);
		}
		
		/* Cancel by user or internal error - both mean stop the presses! */
		if (Abort) {
			fprintf(stderr, "%s: Scan cancelled. %d bytes of image data written to '%s'\n",
				MeName, TotalLen, FileN);
			CmdToIssue[0] = CAN;		/* Cancel the scan */
		        DEBUG("\tSending CAN '%s' to scanner\n", Hex(CmdToIssue, 1));
			DoWrite((char *) &CmdToIssue, 1);
			DoRead(&CmdToIssue, 1) ; 	 /* Get ACK to CAN */
			(void) TIFFClose(File);
			return(0);
		}

		/* Got data, now send ACK telling scanner we got it, but only more data is to come */		if (!(Header.status & Status_AreaEnd)) {
			CmdToIssue[0] = ACK;
		        DEBUG("\tSending ACK '%s' to scanner\n", Hex(CmdToIssue, 1));
			if (!DoWrite((char *) &CmdToIssue, 1)) return(0);
		}

		/* If not debugging, show some progress with a row of dots */
		TotalLen = TotalLen + BytesPerRow;	/* Keep track of # bytes read */
		Progress = Progress + BytesPerRow;	/* Keep track of # bytes read */
		if ((!Debug) && (Progress>=100000)) {
			if (!Quiet) fprintf(stderr,".");
			Progress = 0;
		}

		/*	Got gray-scale data in DataBuffer.  Now write it out. Nothing special to do.  */
		if ((TIFFWriteScanline(File, DataBuffer, Row, (unsigned) 0)) < 0) {
			fprintf(stderr, "%s: Error writing to file '%s'\n", MeName, FileN);
			(void) TIFFClose(File);
			return(0);
		}
	}
	if (!Quiet) fprintf(stderr,"  %d bytes of TIFF image data written to file '%s'\n", TotalLen, FileN);
	(void) TIFFClose(File);
	return(TotalLen);
}

/*
DoScan(FileN, DocN)

Function:		Scan data from scanner, writing results to file FileN. Set name of
			document to DocN. File created is a ppm format stream.

Returns:		Number of bytes read. 0 means the scanfailed
*/
int DoScan(char *FileN, char *DocN)
{
	struct 	Identify_Data_Block 	ScannerIdentifyDB;  /* Get scanner model here */
	struct Condition_Data_Block ScannerConditionDB;  /* Get scanner info here */
	int RC;

	/* Ask the scanner for information about what we are going to scan. This information is
		needed for the header of the file to be created. */
	if (!CmdReqData(C_Request_Condition, &ScannerConditionDB.header)) return(0);

	/* Figure out just what kind of picture this is, then call the scan routine for what works today */
	/* fprintf(stderr, "Color_Mode=%2.2x\n", ScannerConditionDB.Color_Mode); */
	switch (ScannerConditionDB.Color_Mode) {
		case '\001' :
			RC = DoColor24Page(&ScannerConditionDB, FileN, DocN);
			break;
		case '\002' :
			RC = DoColor24Scan(&ScannerConditionDB, FileN, DocN);
			break;
		default :
			RC = DoGrayScan(&ScannerConditionDB, FileN, DocN);
	}
	return(RC);
}

/*
CmdWithParm (Cmd, Parm, Length)

Function:		Issue Cmd and its parameter Parm for length bytes to scanner. Cmd returns no data.

Returns:		0 or 1 depending whether the Cmd failed or not
*/
int CmdWithParm(int Cmd, char *Parm, int Length)
{
	char CmdToIssue[2];	/* Command to send to scanner */

	/* Build the stream to be sent to the device */
	CmdToIssue[0] = ESC;
	CmdToIssue[1] = (char) Cmd;
        DEBUG("\tSending cmd '%s' to scanner\n", Hex(CmdToIssue, 2));
	if (!DoWrite((char *) &CmdToIssue, 2)) return(0);
	
	/* Now read the one byte response - should be ACK or NAK. */
	if (AckNak() != 0) return(0);

        DEBUG("\tSending data %s\n", Hex(Parm, Length));
 	if (!DoWrite(Parm, Length)) return(0);
	/* Now read the one byte response - should be ACK or NAK. */
	return(AckNak() );
}


/*
CmdReqData (Cmd, Data)

Function:		Issue Cmd to scanner. Cmd requires no parms, but returns data in Data.

Returns:		Length of data returned. 0 means the command failed.
			Data from scanner returned in Data.
*/
int CmdReqData(char Cmd, struct IDB *Data)
{
	char CmdToIssue[2];	/* Command to send to scanner */
	int  Len;				/* Number of bytes returned */
	char *P;
	
	/* Build the stream to be sent to the device */
	CmdToIssue[0] = ESC;
	CmdToIssue[1] = Cmd;
        DEBUG("\tSending cmd '%s' to scanner\n", Hex(CmdToIssue, 2));
	if (!DoWrite((char *) &CmdToIssue, 2)) return(0);
	
	/* Now read the data block. First byte is STX, 1 byte of status, 2-byte data length, then data */
	if (!DoRead ( (char *) Data,  sizeof(struct IDB))) return(0);
	if (Debug) {
		if (Data->header == STX)
			fprintf(stderr,"\tInformation block=%s\n", Hex((char *)Data, sizeof(struct IDB)));
	}

	Len = ByteCounter(&(Data->length));
	if (Len == 0) return(4);		/* If no data returned, we are done */
	
	P =  (char *) Data + sizeof(Data);		/* Read next data into here */
	if (!DoRead (P,  Len)) return(0);
	if ((*P!=NAK) && (Debug) ) {
		if (Debug>1) fprintf(stderr,"\tReceived %d bytes=%s\n", Len, Hex(P, Len));
		else fprintf(stderr,"\tReceived %d bytes\n", Len);
	}
	return(Len+4);
}


/*
CmdNoParms (Cmd)

Function:		Issue Cmd to scanner. Cmd requires no parms or data.

Returns:		0 or 1 depending whether the Cmd failed or not

*/
int CmdNoParms(int Cmd)
{
	char CmdToIssue[2];	/* Command to send to scanner */

	/* Build the stream to be sent to the device */
	CmdToIssue[0] = ESC;
	CmdToIssue[1] = (char) Cmd;
        DEBUG("\tSending cmd '%s' to scanner\n", Hex(CmdToIssue, 2));
	if (!DoWrite((char *) &CmdToIssue, 2)) return(1);
	
	/* Now read the one byte response - should be ACK or NAK. */
	return(AckNak());
}

/*
AckNak ()

Function:		Read one byte from scanner, check for Ack or Nak

Returns:		0 if ACK received, 1 otherwise
*/
int AckNak()
{
	char AckNakResponse;		/* Read into here */
	
	if (DoRead(&AckNakResponse, 1) == 0) return (1);    /* pretend NAK if read failed */
	if (AckNakResponse == ACK)  return(0);
	return(1);
}

